//
// Generated by JTB 1.3.2
//
package org.ow2.mind.idl.jtb.visitor;

import org.ow2.mind.idl.jtb.syntaxtree.*;
import java.util.*;

/**
 * A skeleton output formatter for your language grammar.  Using the
 * add() method along with force(), indent(), and outdent(), you can
 * easily specify how this visitor will format the given syntax tree.
 * See the JTB documentation for more details.
 *
 * Pass your syntax tree to this visitor, and then to the TreeDumper
 * visitor in order to "pretty print" your tree.
 */
public class TreeFormatter extends DepthFirstVisitor {
   private Vector<FormatCommand> cmdQueue = new Vector<FormatCommand>();
   private boolean lineWrap;
   private int wrapWidth;
   private int indentAmt;
   private int curLine = 1;
   private int curColumn = 1;
   private int curIndent = 0;

   /**
    * The default constructor assumes an indentation amount of 3 spaces
    * and no line-wrap.  You may alternately use the other constructor to
    * specify your own indentation amount and line width.
    */
   public TreeFormatter() { this(3, 0); }

   /**
    * This constructor accepts an indent amount and a line width which is
    * used to wrap long lines.  If a token's beginColumn value is greater
    * than the specified wrapWidth, it will be moved to the next line and
    * indented one extra level.  To turn off line-wrapping, specify a
    * wrapWidth of 0.
    *
    * @param   indentAmt   Amount of spaces per indentation level.
    * @param   wrapWidth   Wrap lines longer than wrapWidth.  0 for no wrap.
    */
   public TreeFormatter(int indentAmt, int wrapWidth) {
      this.indentAmt = indentAmt;
      this.wrapWidth = wrapWidth;

      if ( wrapWidth > 0 )
         lineWrap = true;
      else
         lineWrap = false;
   }

   /**
    * Accepts a NodeListInterface object and performs an optional format
    * command between each node in the list (but not after the last node).
    */
   protected void processList(NodeListInterface n) {
      processList(n, null);
   }

   protected void processList(NodeListInterface n, FormatCommand cmd) {
      for ( Enumeration<Node> e = n.elements(); e.hasMoreElements(); ) {
         e.nextElement().accept(this);
         if ( cmd != null && e.hasMoreElements() )
            cmdQueue.addElement(cmd);
      }
   }

   /**
    * A Force command inserts a line break and indents the next line to
    * the current indentation level.  Use "add(force());".
    */
   protected FormatCommand force() { return force(1); }
   protected FormatCommand force(int i) {
      return new FormatCommand(FormatCommand.FORCE, i);
   }

   /**
    * An Indent command increases the indentation level by one (or a
    * user-specified amount).  Use "add(indent());".
    */
   protected FormatCommand indent() { return indent(1); }
   protected FormatCommand indent(int i) {
      return new FormatCommand(FormatCommand.INDENT, i);
   }

   /**
    * An Outdent command is the reverse of the Indent command: it reduces
    * the indentation level.  Use "add(outdent());".
    */
   protected FormatCommand outdent() { return outdent(1); }
   protected FormatCommand outdent(int i) {
      return new FormatCommand(FormatCommand.OUTDENT, i);
   }

   /**
    * A Space command simply adds one or a user-specified number of
    * spaces between tokens.  Use "add(space());".
    */
   protected FormatCommand space() { return space(1); }
   protected FormatCommand space(int i) {
      return new FormatCommand(FormatCommand.SPACE, i);
   }

   /**
    * Use this method to add FormatCommands to the command queue to be
    * executed when the next token in the tree is visited.
    */
   protected void add(FormatCommand cmd) {
      cmdQueue.addElement(cmd);
   }

   /**
    * Executes the commands waiting in the command queue, then inserts the
    * proper location information into the current NodeToken.
    *
    * If there are any special tokens preceding this token, they will be
    * given the current location information.  The token will follow on
    * the next line, at the proper indentation level.  If this is not the
    * behavior you want from special tokens, feel free to modify this
    * method.
    */
   public void visit(NodeToken n) {
      for ( Enumeration<FormatCommand> e = cmdQueue.elements(); e.hasMoreElements(); ) {
         FormatCommand cmd = e.nextElement();
         switch ( cmd.getCommand() ) {
         case FormatCommand.FORCE :
            curLine += cmd.getNumCommands();
            curColumn = curIndent + 1;
            break;
         case FormatCommand.INDENT :
            curIndent += indentAmt * cmd.getNumCommands();
            break;
         case FormatCommand.OUTDENT :
            if ( curIndent >= indentAmt )
               curIndent -= indentAmt * cmd.getNumCommands();
            break;
         case FormatCommand.SPACE :
            curColumn += cmd.getNumCommands();
            break;
         default :
            throw new TreeFormatterException(
               "Invalid value in command queue.");
         }
      }

      cmdQueue.removeAllElements();

      //
      // Handle all special tokens preceding this NodeToken
      //
      if ( n.numSpecials() > 0 )
         for ( Enumeration<NodeToken> e = n.specialTokens.elements();
               e.hasMoreElements(); ) {
            NodeToken special = e.nextElement();

            //
            // -Place the token.
            // -Move cursor to next line after the special token.
            // -Don't update curColumn--want to keep current indent level.
            //
            placeToken(special, curLine, curColumn);
            curLine = special.endLine + 1;
         }

      placeToken(n, curLine, curColumn);
      curLine = n.endLine;
      curColumn = n.endColumn;
   }

   /**
    * Inserts token location (beginLine, beginColumn, endLine, endColumn)
    * information into the NodeToken.  Takes into account line-wrap.
    * Does not update curLine and curColumn.
    */
   private void placeToken(NodeToken n, int line, int column) {
      int length = n.tokenImage.length();

      //
      // Find beginning of token.  Only line-wrap for single-line tokens
      //
      if ( !lineWrap || n.tokenImage.indexOf('\n') != -1 ||
           column + length <= wrapWidth )
         n.beginColumn = column;
      else {
         ++line;
         column = curIndent + indentAmt + 1;
         n.beginColumn = column;
      }

      n.beginLine = line;

      //
      // Find end of token; don't count \n if it's the last character
      //
      for ( int i = 0; i < length; ++i ) {
         if ( n.tokenImage.charAt(i) == '\n' && i < length - 1 ) {
            ++line;
            column = 1;
         }
         else
            ++column;
      }

      n.endLine = line;
      n.endColumn = column;
   }

   //
   // User-generated visitor methods below
   //

   /**
    * <PRE>
    * f0 -> ( IncludeDirective() )*
    * f1 -> ( ( TypeDefinition() | ConstantDefinition() ) )*
    * f2 -> InterfaceDefinition()
    * f3 -> [ &lt;SEMICOLON&gt; ]
    * f4 -> &lt;EOF&gt;
    * </PRE>
    */
   public void visit(ITFFile n) {
      if ( n.f0.present() ) {
         processList(n.f0);
      }
      if ( n.f1.present() ) {
         processList(n.f1);
      }
      n.f2.accept(this);
      if ( n.f3.present() ) {
         n.f3.accept(this);
      }
      n.f4.accept(this);
   }

   /**
    * <PRE>
    * f0 -> &lt;PREPROC_START&gt;
    * f1 -> &lt;PREPROC_IFNDEF&gt;
    * f2 -> &lt;PREPROC_IDENTIFIER&gt;
    * f3 -> &lt;PREPROC_ROL&gt;
    * f4 -> &lt;PREPROC_START&gt;
    * f5 -> &lt;PREPROC_DEFINE&gt;
    * f6 -> &lt;PREPROC_IDENTIFIER&gt;
    * f7 -> &lt;PREPROC_ROL&gt;
    * f8 -> ( IncludeDirective() )*
    * f9 -> ( ( TypeDefinition() | ConstantDefinition() ) )*
    * f10 -> &lt;PREPROC_START&gt;
    * f11 -> &lt;PREPROC_ENDIF&gt;
    * f12 -> [ &lt;PREPROC_ROL&gt; ]
    * f13 -> &lt;EOF&gt;
    * </PRE>
    */
   public void visit(IDTFile n) {
      n.f0.accept(this);
      n.f1.accept(this);
      n.f2.accept(this);
      n.f3.accept(this);
      n.f4.accept(this);
      n.f5.accept(this);
      n.f6.accept(this);
      n.f7.accept(this);
      if ( n.f8.present() ) {
         processList(n.f8);
      }
      if ( n.f9.present() ) {
         processList(n.f9);
      }
      n.f10.accept(this);
      n.f11.accept(this);
      if ( n.f12.present() ) {
         n.f12.accept(this);
      }
      n.f13.accept(this);
   }

   /**
    * <PRE>
    * f0 -> &lt;PREPROC_START&gt;
    * f1 -> &lt;PREPROC_INCLUDE&gt;
    * f2 -> &lt;PREPROC_PATH&gt;
    * f3 -> &lt;PREPROC_ROL&gt;
    * </PRE>
    */
   public void visit(IncludeDirective n) {
      n.f0.accept(this);
      n.f1.accept(this);
      n.f2.accept(this);
      n.f3.accept(this);
   }

   /**
    * <PRE>
    * f0 -> &lt;PREPROC_START&gt;
    * f1 -> &lt;PREPROC_DEFINE&gt;
    * f2 -> &lt;PREPROC_IDENTIFIER&gt;
    * f3 -> &lt;PREPROC_ROL&gt;
    * </PRE>
    */
   public void visit(ConstantDefinition n) {
      n.f0.accept(this);
      n.f1.accept(this);
      n.f2.accept(this);
      n.f3.accept(this);
   }

   /**
    * <PRE>
    * f0 -> ( TypeDefSpecification() | StructOrUnionSpecification() | EnumSpecification() )
    * f1 -> &lt;SEMICOLON&gt;
    * </PRE>
    */
   public void visit(TypeDefinition n) {
      n.f0.accept(this);
      n.f1.accept(this);
   }

   /**
    * <PRE>
    * f0 -> &lt;TYPEDEF&gt;
    * f1 -> QualifiedTypeSpecification()
    * f2 -> Declarators()
    * </PRE>
    */
   public void visit(TypeDefSpecification n) {
      n.f0.accept(this);
      n.f1.accept(this);
      n.f2.accept(this);
   }

   /**
    * <PRE>
    * f0 -> ( TypeQualifier() )*
    * f1 -> TypeSpecification()
    * </PRE>
    */
   public void visit(QualifiedTypeSpecification n) {
      if ( n.f0.present() ) {
         processList(n.f0);
      }
      n.f1.accept(this);
   }

   /**
    * <PRE>
    * f0 -> &lt;CONST&gt;
    *       | &lt;VOLATILE&gt;
    * </PRE>
    */
   public void visit(TypeQualifier n) {
      n.f0.accept(this);
   }

   /**
    * <PRE>
    * f0 -> TypeDefName()
    *       | StructOrUnionSpecification()
    *       | EnumSpecification()
    *       | TypeSpecifiers()
    * </PRE>
    */
   public void visit(TypeSpecification n) {
      n.f0.accept(this);
   }

   /**
    * <PRE>
    * f0 -> FullyQualifiedName()
    * </PRE>
    */
   public void visit(TypeDefName n) {
      n.f0.accept(this);
   }

   /**
    * <PRE>
    * f0 -> ( &lt;VOID&gt; | &lt;CHAR&gt; | &lt;SHORT&gt; | &lt;INT&gt; | &lt;LONG&gt; | &lt;FLOAT&gt; | &lt;DOUBLE&gt; | &lt;SIGNED&gt; | &lt;UNSIGNED&gt; | &lt;STRING&gt; | &lt;INT8_T&gt; | &lt;UINT8_T&gt; | &lt;INT16_T&gt; | &lt;UINT16_T&gt; | &lt;INT32_T&gt; | &lt;UINT32_T&gt; | &lt;INT64_T&gt; | &lt;UINT64_T&gt; | &lt;INTPTR_T&gt; | &lt;UINTPTR_T&gt; )+
    * </PRE>
    */
   public void visit(TypeSpecifiers n) {
      processList(n.f0);
   }

   /**
    * <PRE>
    * f0 -> StructOrUnionDefinition()
    *       | StructOrUnionReference()
    * </PRE>
    */
   public void visit(StructOrUnionSpecification n) {
      n.f0.accept(this);
   }

   /**
    * <PRE>
    * f0 -> StructOrUnion()
    * f1 -> [ &lt;IDENTIFIER&gt; ]
    * f2 -> &lt;LBRACE&gt;
    * f3 -> StructMemberList()
    * f4 -> &lt;RBRACE&gt;
    * </PRE>
    */
   public void visit(StructOrUnionDefinition n) {
      n.f0.accept(this);
      if ( n.f1.present() ) {
         n.f1.accept(this);
      }
      n.f2.accept(this);
      n.f3.accept(this);
      n.f4.accept(this);
   }

   /**
    * <PRE>
    * f0 -> StructOrUnion()
    * f1 -> &lt;IDENTIFIER&gt;
    * </PRE>
    */
   public void visit(StructOrUnionReference n) {
      n.f0.accept(this);
      n.f1.accept(this);
   }

   /**
    * <PRE>
    * f0 -> &lt;STRUCT&gt;
    *       | &lt;UNION&gt;
    * </PRE>
    */
   public void visit(StructOrUnion n) {
      n.f0.accept(this);
   }

   /**
    * <PRE>
    * f0 -> ( StructMember() )*
    * </PRE>
    */
   public void visit(StructMemberList n) {
      if ( n.f0.present() ) {
         processList(n.f0);
      }
   }

   /**
    * <PRE>
    * f0 -> Annotations()
    * f1 -> QualifiedTypeSpecification()
    * f2 -> Declarators()
    * f3 -> [ &lt;COLON&gt; ConstantExpression() ]
    * f4 -> &lt;SEMICOLON&gt;
    * </PRE>
    */
   public void visit(StructMember n) {
      n.f0.accept(this);
      n.f1.accept(this);
      n.f2.accept(this);
      if ( n.f3.present() ) {
         n.f3.accept(this);
      }
      n.f4.accept(this);
   }

   /**
    * <PRE>
    * f0 -> EnumDefinition()
    *       | EnumReference()
    * </PRE>
    */
   public void visit(EnumSpecification n) {
      n.f0.accept(this);
   }

   /**
    * <PRE>
    * f0 -> &lt;ENUM&gt;
    * f1 -> [ &lt;IDENTIFIER&gt; ]
    * f2 -> &lt;LBRACE&gt;
    * f3 -> EnumMemberList()
    * f4 -> &lt;RBRACE&gt;
    * </PRE>
    */
   public void visit(EnumDefinition n) {
      n.f0.accept(this);
      if ( n.f1.present() ) {
         n.f1.accept(this);
      }
      n.f2.accept(this);
      n.f3.accept(this);
      n.f4.accept(this);
   }

   /**
    * <PRE>
    * f0 -> &lt;ENUM&gt;
    * f1 -> &lt;IDENTIFIER&gt;
    * </PRE>
    */
   public void visit(EnumReference n) {
      n.f0.accept(this);
      n.f1.accept(this);
   }

   /**
    * <PRE>
    * f0 -> EnumMember()
    * f1 -> ( &lt;COMMA&gt; EnumMember() )*
    * </PRE>
    */
   public void visit(EnumMemberList n) {
      n.f0.accept(this);
      if ( n.f1.present() ) {
         processList(n.f1);
      }
   }

   /**
    * <PRE>
    * f0 -> Annotations()
    * f1 -> &lt;IDENTIFIER&gt;
    * f2 -> [ EnumValue() ]
    * </PRE>
    */
   public void visit(EnumMember n) {
      n.f0.accept(this);
      n.f1.accept(this);
      if ( n.f2.present() ) {
         n.f2.accept(this);
      }
   }

   /**
    * <PRE>
    * f0 -> &lt;ASSIGN&gt;
    * f1 -> ConstantExpression()
    * </PRE>
    */
   public void visit(EnumValue n) {
      n.f0.accept(this);
      n.f1.accept(this);
   }

   /**
    * <PRE>
    * f0 -> Declarator()
    * f1 -> ( &lt;COMMA&gt; Declarator() )*
    * </PRE>
    */
   public void visit(Declarators n) {
      n.f0.accept(this);
      if ( n.f1.present() ) {
         processList(n.f1);
      }
   }

   /**
    * <PRE>
    * f0 -> PointerSpecification()
    * f1 -> DirectDeclarator()
    * </PRE>
    */
   public void visit(Declarator n) {
      n.f0.accept(this);
      n.f1.accept(this);
   }

   /**
    * <PRE>
    * f0 -> ( QualifierPointerSpecification() )*
    * </PRE>
    */
   public void visit(PointerSpecification n) {
      if ( n.f0.present() ) {
         processList(n.f0);
      }
   }

   /**
    * <PRE>
    * f0 -> &lt;STAR&gt;
    * f1 -> ( TypeQualifier() )*
    * </PRE>
    */
   public void visit(QualifierPointerSpecification n) {
      n.f0.accept(this);
      if ( n.f1.present() ) {
         processList(n.f1);
      }
   }

   /**
    * <PRE>
    * f0 -> ( &lt;IDENTIFIER&gt; | &lt;LPAREN&gt; Declarator() &lt;RPAREN&gt; )
    * f1 -> ( ArraySpecification() )*
    * </PRE>
    */
   public void visit(DirectDeclarator n) {
      n.f0.accept(this);
      if ( n.f1.present() ) {
         processList(n.f1);
      }
   }

   /**
    * <PRE>
    * f0 -> &lt;LBRACKET&gt;
    * f1 -> [ ConstantExpression() ]
    * f2 -> &lt;RBRACKET&gt;
    * </PRE>
    */
   public void visit(ArraySpecification n) {
      n.f0.accept(this);
      if ( n.f1.present() ) {
         n.f1.accept(this);
      }
      n.f2.accept(this);
   }

   /**
    * <PRE>
    * f0 -> PointerSpecification()
    * f1 -> [ AbstractDirectDeclarator() ]
    * </PRE>
    */
   public void visit(AbstractDeclarator n) {
      n.f0.accept(this);
      if ( n.f1.present() ) {
         n.f1.accept(this);
      }
   }

   /**
    * <PRE>
    * f0 -> &lt;LPAREN&gt; AbstractDeclarator() &lt;RPAREN&gt; ( ArraySpecification() )*
    *       | ( ArraySpecification() )+
    * </PRE>
    */
   public void visit(AbstractDirectDeclarator n) {
      n.f0.accept(this);
   }

   /**
    * <PRE>
    * f0 -> Annotations()
    * f1 -> &lt;INTERFACE&gt;
    * f2 -> [ &lt;UNMANAGED&gt; ]
    * f3 -> FullyQualifiedName()
    * f4 -> [ InterfaceInheritanceSpecification() ]
    * f5 -> InterfaceBody()
    * </PRE>
    */
   public void visit(InterfaceDefinition n) {
      n.f0.accept(this);
      n.f1.accept(this);
      if ( n.f2.present() ) {
         n.f2.accept(this);
      }
      n.f3.accept(this);
      if ( n.f4.present() ) {
         n.f4.accept(this);
      }
      n.f5.accept(this);
   }

   /**
    * <PRE>
    * f0 -> &lt;COLON&gt;
    * f1 -> FullyQualifiedName()
    * </PRE>
    */
   public void visit(InterfaceInheritanceSpecification n) {
      n.f0.accept(this);
      n.f1.accept(this);
   }

   /**
    * <PRE>
    * f0 -> &lt;LBRACE&gt;
    * f1 -> ( MethodDefinition() &lt;SEMICOLON&gt; )*
    * f2 -> &lt;RBRACE&gt;
    * </PRE>
    */
   public void visit(InterfaceBody n) {
      n.f0.accept(this);
      if ( n.f1.present() ) {
         processList(n.f1);
      }
      n.f2.accept(this);
   }

   /**
    * <PRE>
    * f0 -> Annotations()
    * f1 -> QualifiedTypeSpecification()
    * f2 -> PointerSpecification()
    * f3 -> &lt;IDENTIFIER&gt;
    * f4 -> Parameters()
    * </PRE>
    */
   public void visit(MethodDefinition n) {
      n.f0.accept(this);
      n.f1.accept(this);
      n.f2.accept(this);
      n.f3.accept(this);
      n.f4.accept(this);
   }

   /**
    * <PRE>
    * f0 -> &lt;LPAREN&gt; [ &lt;VOID&gt; ] &lt;RPAREN&gt;
    *       | &lt;LPAREN&gt; ParameterList() &lt;RPAREN&gt;
    * </PRE>
    */
   public void visit(Parameters n) {
      n.f0.accept(this);
   }

   /**
    * <PRE>
    * f0 -> Parameter()
    * f1 -> ( &lt;COMMA&gt; Parameter() )*
    * f2 -> [ &lt;COMMA&gt; &lt;LDOTS&gt; ]
    * </PRE>
    */
   public void visit(ParameterList n) {
      n.f0.accept(this);
      if ( n.f1.present() ) {
         processList(n.f1);
      }
      if ( n.f2.present() ) {
         n.f2.accept(this);
      }
   }

   /**
    * <PRE>
    * f0 -> Annotations()
    * f1 -> ( ParameterQualifier() )*
    * f2 -> QualifiedTypeSpecification()
    * f3 -> ( Declarator() | AbstractDeclarator() )
    * </PRE>
    */
   public void visit(Parameter n) {
      n.f0.accept(this);
      if ( n.f1.present() ) {
         processList(n.f1);
      }
      n.f2.accept(this);
      n.f3.accept(this);
   }

   /**
    * <PRE>
    * f0 -> &lt;IN&gt;
    *       | &lt;OUT&gt;
    * </PRE>
    */
   public void visit(ParameterQualifier n) {
      n.f0.accept(this);
   }

   /**
    * <PRE>
    * f0 -> LogicalOrExpression()
    * </PRE>
    */
   public void visit(ConstantExpression n) {
      n.f0.accept(this);
   }

   /**
    * <PRE>
    * f0 -> LogicalAndExpression()
    * f1 -> [ "||" LogicalOrExpression() ]
    * </PRE>
    */
   public void visit(LogicalOrExpression n) {
      n.f0.accept(this);
      if ( n.f1.present() ) {
         n.f1.accept(this);
      }
   }

   /**
    * <PRE>
    * f0 -> OrExpression()
    * f1 -> [ "&&" LogicalAndExpression() ]
    * </PRE>
    */
   public void visit(LogicalAndExpression n) {
      n.f0.accept(this);
      if ( n.f1.present() ) {
         n.f1.accept(this);
      }
   }

   /**
    * <PRE>
    * f0 -> XorExpression()
    * f1 -> [ "|" OrExpression() ]
    * </PRE>
    */
   public void visit(OrExpression n) {
      n.f0.accept(this);
      if ( n.f1.present() ) {
         n.f1.accept(this);
      }
   }

   /**
    * <PRE>
    * f0 -> AndExpression()
    * f1 -> [ "^" XorExpression() ]
    * </PRE>
    */
   public void visit(XorExpression n) {
      n.f0.accept(this);
      if ( n.f1.present() ) {
         n.f1.accept(this);
      }
   }

   /**
    * <PRE>
    * f0 -> ShiftExpression()
    * f1 -> [ "&" AndExpression() ]
    * </PRE>
    */
   public void visit(AndExpression n) {
      n.f0.accept(this);
      if ( n.f1.present() ) {
         n.f1.accept(this);
      }
   }

   /**
    * <PRE>
    * f0 -> AdditiveExpression()
    * f1 -> [ ShiftOperation() ShiftExpression() ]
    * </PRE>
    */
   public void visit(ShiftExpression n) {
      n.f0.accept(this);
      if ( n.f1.present() ) {
         n.f1.accept(this);
      }
   }

   /**
    * <PRE>
    * f0 -> MulExpression()
    * f1 -> [ AdditiveOperation() AdditiveExpression() ]
    * </PRE>
    */
   public void visit(AdditiveExpression n) {
      n.f0.accept(this);
      if ( n.f1.present() ) {
         n.f1.accept(this);
      }
   }

   /**
    * <PRE>
    * f0 -> CastExpression()
    * f1 -> [ MulOperation() MulExpression() ]
    * </PRE>
    */
   public void visit(MulExpression n) {
      n.f0.accept(this);
      if ( n.f1.present() ) {
         n.f1.accept(this);
      }
   }

   /**
    * <PRE>
    * f0 -> &lt;LPAREN&gt; QualifiedTypeSpecification() AbstractDeclarator() &lt;RPAREN&gt; CastExpression()
    *       | UnaryExpression()
    * </PRE>
    */
   public void visit(CastExpression n) {
      n.f0.accept(this);
   }

   /**
    * <PRE>
    * f0 -> [ UnaryOperation() ]
    * f1 -> PrimaryExpression()
    * </PRE>
    */
   public void visit(UnaryExpression n) {
      if ( n.f0.present() ) {
         n.f0.accept(this);
      }
      n.f1.accept(this);
   }

   /**
    * <PRE>
    * f0 -> Literal()
    *       | &lt;LPAREN&gt; ConstantExpression() &lt;RPAREN&gt;
    * </PRE>
    */
   public void visit(PrimaryExpression n) {
      n.f0.accept(this);
   }

   /**
    * <PRE>
    * f0 -> &lt;INTEGER_LITERAL&gt;
    *       | &lt;STRING_LITERAL&gt;
    *       | &lt;CHARACTER_LITERAL&gt;
    *       | &lt;FLOATING_POINT_LITERAL&gt;
    *       | &lt;IDENTIFIER&gt;
    * </PRE>
    */
   public void visit(Literal n) {
      n.f0.accept(this);
   }

   /**
    * <PRE>
    * f0 -> &lt;LSHIFT&gt;
    *       | &lt;RSHIFT&gt;
    * </PRE>
    */
   public void visit(ShiftOperation n) {
      n.f0.accept(this);
   }

   /**
    * <PRE>
    * f0 -> &lt;PLUS&gt;
    *       | &lt;MINUS&gt;
    * </PRE>
    */
   public void visit(AdditiveOperation n) {
      n.f0.accept(this);
   }

   /**
    * <PRE>
    * f0 -> &lt;STAR&gt;
    *       | &lt;SLASH&gt;
    *       | &lt;PERCENT&gt;
    * </PRE>
    */
   public void visit(MulOperation n) {
      n.f0.accept(this);
   }

   /**
    * <PRE>
    * f0 -> &lt;AMP&gt;
    *       | &lt;STAR&gt;
    *       | &lt;PLUS&gt;
    *       | &lt;MINUS&gt;
    *       | &lt;TILDE&gt;
    *       | &lt;NOT&gt;
    * </PRE>
    */
   public void visit(UnaryOperation n) {
      n.f0.accept(this);
   }

   /**
    * <PRE>
    * f0 -> ( Annotation() )*
    * </PRE>
    */
   public void visit(Annotations n) {
      if ( n.f0.present() ) {
         processList(n.f0);
      }
   }

   /**
    * <PRE>
    * f0 -> &lt;AT&gt;
    * f1 -> FullyQualifiedName()
    * f2 -> [ AnnotationParameters() ]
    * </PRE>
    */
   public void visit(Annotation n) {
      n.f0.accept(this);
      n.f1.accept(this);
      if ( n.f2.present() ) {
         n.f2.accept(this);
      }
   }

   /**
    * <PRE>
    * f0 -> &lt;LPAREN&gt;
    * f1 -> [ AnnotationValuePairs() | AnnotationValue() ]
    * f2 -> &lt;RPAREN&gt;
    * </PRE>
    */
   public void visit(AnnotationParameters n) {
      n.f0.accept(this);
      if ( n.f1.present() ) {
         n.f1.accept(this);
      }
      n.f2.accept(this);
   }

   /**
    * <PRE>
    * f0 -> AnnotationValuePair()
    * f1 -> ( &lt;COMMA&gt; AnnotationValuePair() )*
    * </PRE>
    */
   public void visit(AnnotationValuePairs n) {
      n.f0.accept(this);
      if ( n.f1.present() ) {
         processList(n.f1);
      }
   }

   /**
    * <PRE>
    * f0 -> &lt;IDENTIFIER&gt;
    * f1 -> &lt;ASSIGN&gt;
    * f2 -> AnnotationValue()
    * </PRE>
    */
   public void visit(AnnotationValuePair n) {
      n.f0.accept(this);
      n.f1.accept(this);
      n.f2.accept(this);
   }

   /**
    * <PRE>
    * f0 -> StringValue()
    *       | IntegerValue()
    *       | BooleanValue()
    *       | AnnotationAnnotationValue()
    *       | ArrayAnnotationValue()
    *       | NullValue()
    * </PRE>
    */
   public void visit(AnnotationValue n) {
      n.f0.accept(this);
   }

   /**
    * <PRE>
    * f0 -> Annotation()
    * </PRE>
    */
   public void visit(AnnotationAnnotationValue n) {
      n.f0.accept(this);
   }

   /**
    * <PRE>
    * f0 -> &lt;LBRACE&gt;
    * f1 -> [ AnnotationValue() ( &lt;COMMA&gt; AnnotationValue() )* ]
    * f2 -> &lt;RBRACE&gt;
    * </PRE>
    */
   public void visit(ArrayAnnotationValue n) {
      n.f0.accept(this);
      if ( n.f1.present() ) {
         n.f1.accept(this);
      }
      n.f2.accept(this);
   }

   /**
    * <PRE>
    * f0 -> &lt;STRING_LITERAL&gt;
    * </PRE>
    */
   public void visit(StringValue n) {
      n.f0.accept(this);
   }

   /**
    * <PRE>
    * f0 -> [ &lt;MINUS&gt; | &lt;PLUS&gt; ]
    * f1 -> &lt;INTEGER_LITERAL&gt;
    * </PRE>
    */
   public void visit(IntegerValue n) {
      if ( n.f0.present() ) {
         n.f0.accept(this);
      }
      n.f1.accept(this);
   }

   /**
    * <PRE>
    * f0 -> &lt;BOOLEAN_LITERAL&gt;
    * </PRE>
    */
   public void visit(BooleanValue n) {
      n.f0.accept(this);
   }

   /**
    * <PRE>
    * f0 -> &lt;NULL&gt;
    * </PRE>
    */
   public void visit(NullValue n) {
      n.f0.accept(this);
   }

   /**
    * <PRE>
    * f0 -> &lt;IDENTIFIER&gt;
    * f1 -> ( &lt;DOT&gt; &lt;IDENTIFIER&gt; )*
    * </PRE>
    */
   public void visit(FullyQualifiedName n) {
      n.f0.accept(this);
      if ( n.f1.present() ) {
         processList(n.f1);
      }
   }

}

class FormatCommand {
   public static final int FORCE = 0;
   public static final int INDENT = 1;
   public static final int OUTDENT = 2;
   public static final int SPACE = 3;

   private int command;
   private int numCommands;

   FormatCommand(int command, int numCommands) {
      this.command = command;
      this.numCommands = numCommands;
   }

   public int getCommand()             { return command; }
   public int getNumCommands()         { return numCommands; }
   public void setCommand(int i)       { command = i; }
   public void setNumCommands(int i)   { numCommands = i; }
}

class TreeFormatterException extends RuntimeException {
   TreeFormatterException()         { super(); }
   TreeFormatterException(String s) { super(s); }
}
